#!/usr/bin/env python
"""
Copyright (c) 2009, Patrick Salecker
All rights reserved.

Redistribution and use in source and binary forms, with or without 
modification, are permitted provided that the following conditions are met:

    * Redistributions of source code must retain the above copyright notice, 
      this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice, 
      this list of conditions and the following disclaimer in 
      the documentation and/or other materials provided with the distribution.
    * Neither the name of the author nor the names of its 
      contributors may be used to endorse or promote products derived 
      from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, 
THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS 
BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, 
OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
POSSIBILITY OF SUCH DAMAGE.

"""
__version__="0.3"

import os
import time
import sys
import math

import gtk
import gobject

try:
	import gps
except ImportError:
	error="%s\nTry install python-gps!"%sys.exc_info()[1]
	gtk.MessageDialog(None, gtk.DIALOG_DESTROY_WITH_PARENT, gtk.MESSAGE_ERROR, gtk.BUTTONS_CLOSE, error).run()
	print error,"\nExit!"
	sys.exit()

class GGPS(gtk.StatusIcon):
	def __init__(self):
		self.name="GGPS"
		gtk.StatusIcon.__init__(self)
		menu = '''
			<ui>
				<menubar name="Menubar">
					<menu action="Menu">
						<menuitem action="Reconnect"/>
						<separator/>
						<menuitem action="About"/>
						<menuitem action="Close"/>
					</menu>
				</menubar>
			</ui>'''
		
		actions = [
			('Menu',  None, 'Menu'),
			('Reconnect', gtk.STOCK_REFRESH, '_Reconnect', None, 'Reconnect to GPSD', self.gps_reconnect),
			('About', gtk.STOCK_ABOUT, '_About', None, 'About %s'%self.name, self.on_about),
			('Close', gtk.STOCK_QUIT, '_Close', None, 'Quit', self.quit)
			]
		ag = gtk.ActionGroup('Actions')
		ag.add_actions(actions)
		self.manager = gtk.UIManager()
		self.manager.insert_action_group(ag, 0)
		self.manager.add_ui_from_string(menu)
		self.menu = self.manager.get_widget('/Menubar/Menu/About').props.parent
		
		self.set_tooltip(self.name)
		self.set_visible(True)
		self.connect('activate', self.on_activate)
		self.connect('popup-menu', self.on_popup_menu)

		self.sats=[]

		self.status=0
		self.status_timer=time.time()
		
		self.new_window()
		self.win.show()
		
		self.gps_con=GpsConnection()
		self.gps_connect()
		
		gobject.timeout_add(1000, self.updater)
		
	def set_icon(self,status=None):
		if status==None:
			self.set_from_stock(gtk.STOCK_DISCONNECT)
		elif status=="connected":
			self.set_from_stock(gtk.STOCK_CONNECT)
		elif status=="disconnected":
			self.set_from_stock(gtk.STOCK_DISCONNECT)
		elif status=="fix":
			self.set_from_stock(gtk.STOCK_YES)
		elif status=="nofix":
			self.set_from_stock(gtk.STOCK_NO)
		
	def on_activate(self, data):
		"""show/hide the window on click
		"""
		if self.win.is_visible==False:
			if self.win.gtkwin==None:
				self.new_window()
			self.win.show()
		else:
			self.win.hide()

	def on_popup_menu(self, status, button, time):
		self.menu.popup(None, None, None, button, time)

	def on_about(self, widget):
		"""open the about dialog
		"""
		dialog = gtk.AboutDialog()
		dialog.set_name(self.name)
		dialog.set_version(__version__)
		dialog.set_comments('A GPSD Client that shows the GPS status in the tray')
		#dialog.set_website('')
		dialog.set_copyright("(c) 2009 Patrick Salecker")
		dialog.run()
		dialog.destroy()

	def new_window(self):
		"""create a new window and its structure
		"""
		self.win=Window(self.name,self.gps_reconnect,self.status_timer)
		
	def gps_connect(self):
		# connect to gpsd
		self.set_status(0)
		if self.gps_con.connect():
			self.set_icon("connected")
		else:
			self.set_icon("disconnected")
	
	def gps_reconnect(self,widget=None,data=None):
		print "reconnect"
		self.gps_connect()
	
	def quit(self,data=None):
		gtk.main_quit()
		
	def set_status(self,status):
		if status!=self.status:
			self.status_timer=time.time()
			self.win.status_timer=self.status_timer
			self.status=status
		
	def updater(self):
		# called every second
		if self.gps_con.update()==True:
			data=self.gps_con.get_data()
			if data["status"]==1 and data["time"]>0:
				self.set_icon("fix")
			else:
				self.set_icon("nofix")
			
			self.set_status(data["status"])
			
			if self.win.is_visible==True:
				self.win.fill(data)
		else:
			self.gps_reconnect()
			
		return True
		
class Window:
	def __init__(self,name,gps_reconnect,status_timer):
		self.gps_reconnect=gps_reconnect
		self.status_timer=status_timer
		self.gtkwin=gtk.Window(gtk.WINDOW_TOPLEVEL)
		self.gtkwin.connect("destroy", self.destroy)
		self.gtkwin.set_border_width(5)
		self.gtkwin.set_position(gtk.WIN_POS_CENTER)
		self.gtkwin.set_title("%s %s" %(name,__version__))
		
		self.is_fullscreen=False
		
		self.gtkwin.connect('key-press-event',self.key_press)
		
		self.sat_scrolled_window = gtk.ScrolledWindow()
		self.sat_scrolled_window.set_policy(gtk.POLICY_NEVER, gtk.POLICY_AUTOMATIC)
		self.sat_scrolled_window.set_shadow_type(gtk.SHADOW_NONE)
		self.sat_scrolled_window.show()
		
		self.sat_frame = gtk.Frame()
		self.sat_frame.add(self.sat_scrolled_window)
		self.sat_frame.show()
		
		self.gps_frame = gtk.Frame("GPS Data")
		self.gps_frame.show()
		
		button = gtk.Button("Reconnect")
		button.connect("clicked", self.gps_reconnect, None)
		button.show()
		
		button_box=gtk.HButtonBox()
		button_box.show()
		button_box.add(button)
		
		self.left_table=gtk.Table(rows=3, columns=1)
		self.left_table.show()
		self.left_table.attach(self.sat_frame, 0, 1, 0, 1)
		self.left_table.attach(self.gps_frame,0, 1, 1, 2,  yoptions=gtk.SHRINK)
		self.left_table.attach(button_box, 0, 1, 2, 3, yoptions=gtk.SHRINK)
		
		alignment=gtk.Alignment(0,0,0,1)
		alignment.add(self.left_table)
		alignment.show()
		
		self.hbox=gtk.HBox()
		self.hbox.show()
		self.hbox.pack_start(alignment, expand=False, fill=False, padding=0)
		
		self.gtkwin.add(self.hbox)
		
		self.is_visible=False
		
	def destroy(self,widget):
		pass
		self.is_visible=False
		self.gtkwin=None
	
	def show(self,widget=None):
		print "show window"
		self.gtkwin.show()
		self.is_visible=True
	
	def hide(self,widget=None):
		print "hide window"
		self.gtkwin.hide()
		self.is_visible=False
		
	def fill(self,data):
		if len(self.sat_scrolled_window.get_children())>0:
			self.sat_scrolled_window.remove(self.treeview)
			self.gps_frame.remove(self.gpstable)
			self.hbox.remove(self.skyview_alignment)
			
		try:
			loctime=time.localtime(data["time"])
			if loctime>0:
				data["time"]=time.strftime("%d.%m.%y %H:%M:%S",loctime)
		except:
			#print sys.exc_info()
			pass
			
		if data["status"]==0 or data["time"]==0:
			data["status"]="No FIX"
			if data["time"]==0 and data["status"]!=0:
				data["status"]+=" (?!)"
		else:
			data["status"]="FIX"
		data["status"]+=" (%s sec)"%int(time.time()-self.status_timer)
			
		self.sats=data["satellites"]
		self.sat_list()
		self.gps_data_table(data)
		self.create_skyview()
		
	def sat_list(self):
		"""a list that contains information about the satellites
		"""
		
		self.treestore = gtk.TreeStore(
			gobject.TYPE_STRING,
			gobject.TYPE_STRING,
			gobject.TYPE_STRING,
			gobject.TYPE_STRING,
			gobject.TYPE_STRING
			)
		
		satsused=0
		for s in self.sats:
			if s.used==1:
				used="Yes"
			else:
				used="No"
			row=[s.PRN,s.azimuth,s.elevation,s.ss,used]
			piter = self.treestore.append(None, row)
			if s.used==1:
				satsused+=1
		
		self.treeview = gtk.TreeView(self.treestore)
		columns=("PRN","Azimuth","Elevation","SNR","Used")
		num=0
		for column in columns:
			tvcolumn = gtk.TreeViewColumn(column)
			self.treeview.append_column(tvcolumn)
			cell = gtk.CellRendererText()
			tvcolumn.pack_start(cell, True)
			tvcolumn.add_attribute(cell, 'text', num)
			#tvcolumn.set_sort_column_id(num)
			num+=1
		#self.treeview.set_search_column(0)
		self.treeview.set_reorderable(True)
		
		self.sat_frame.set_label("Satellites (%s of %s used)" % (satsused,len(self.sats)))
		
		self.sat_scrolled_window.add(self.treeview)
		self.treeview.show()
	
	def gps_data_table(self,data):
		"""a table that shows some GPS data
		"""
		
		# titles for the rows
		gpsrows=(
			("Latitude","latitude"),
			("Longitude","longitude"),
			("Altitude","altitude"),
			("Speed","speed"),
			("Time","time"),
			("UTC","utc"),
			("Fix","status")
			)
		
		self.gpstable=gtk.Table(len(gpsrows),2)
		row=0
		for title,varname in gpsrows:
			value=data[varname]
			label = gtk.Label("%s:"%title)
			label.set_alignment(xalign=0, yalign=0)
			self.gpstable.attach(label, 0, 1, row, row+1)
			label.show()
			
			label = gtk.Label(value)
			label.set_alignment(xalign=0, yalign=0)
			self.gpstable.attach(label, 1, 2, row, row+1)
			label.show()
			row+=1
		
		self.gps_frame.add(self.gpstable)
		self.gpstable.show()
		
	def create_skyview(self):
		self.skyview=gtk.DrawingArea()
		self.skyview.connect("expose_event", self.draw_skyview)
		self.skyview.set_size_request(300, 300)
		
		self.skyview_frame = gtk.Frame("Skyview")
		self.skyview_frame.add(self.skyview)
		self.skyview_frame.show()
		
		self.skyview_alignment=gtk.Alignment(0,0,1,1)
		self.skyview_alignment.add(self.skyview_frame)
		self.skyview_alignment.show()
		
		self.hbox.pack_end(self.skyview_alignment, expand=True, fill=True, padding=0)
		self.skyview.show()
		
	def draw_skyview(self,widget,event):
		ctx=widget.window.cairo_create()
		
		size=min(event.area.width,event.area.height)
		mid=size/2
		radius=mid-15
		
		# white background
		ctx.set_source_rgb(1,1,1)
		ctx.rectangle(0, 0,event.area.width,event.area.height)
		ctx.fill()
		ctx.stroke()
		
		ctx.set_line_width(1)
		ctx.set_source_rgb(0, 0, 0)
		
		# 3 Arcs
		for arcrad in (radius, radius/2, 5):
			ctx.arc(mid,mid, arcrad, 0, 2.0 * math.pi)
			ctx.stroke()
				
		ctx.set_source_rgba(0,0,0)
		
		# North,South,East,West labels
		ctx.move_to(mid-3,12)
		ctx.show_text("N")
		ctx.move_to(mid-3,size-5)
		ctx.show_text("S")
		ctx.move_to(size-12,mid)
		ctx.show_text("E")
		ctx.move_to(3,mid)
		ctx.show_text("W")
		ctx.stroke()
		
		for s in self.sats:
			#s.PRN,s.azimuth,s.elevation,s.ss,s.used
			# chose color
			if s.ss<30:
				ctx.set_source_rgb(1, 0, 0) #red
			elif s.ss<35:
				ctx.set_source_rgb(1,1, 0) #yellow
			elif s.ss<40:
				ctx.set_source_rgba(0,0.6, 0) #darkgreen
			else:
				ctx.set_source_rgb(0, 1, 0) #green
			
			# elev = 0-90
			# azi = 0-360
			foo=radius-(1.0/90.0*s.elevation*radius)
			bla=(s.azimuth-90)*math.pi/180
			x=mid+foo*math.cos(bla)
			y=mid+foo*math.sin(bla)
			ctx.arc(int(x),int(y), 6, 0, 2.0*math.pi)
			
			# fill if sat is used
			if s.used==1:
				ctx.fill()
			else:
				ctx.set_line_width(1)
			ctx.stroke()
			
			# Sat PRN label
			ctx.move_to(x+5,y+12)
			ctx.set_source_rgba(0,0,0)
			ctx.show_text(str(s.PRN))
			ctx.stroke()
			
		return False
		
	def key_press(self,widget,event):
		keyval = event.keyval
		name = gtk.gdk.keyval_name(keyval)
		#mod = gtk.accelerator_get_label(keyval,event.state)
		if name=="F11":
			print "Key:",name
			self.fullscreen()
			
	def fullscreen(self):
		if self.is_fullscreen==True:
			self.gtkwin.unfullscreen()
			self.is_fullscreen=False
		else:
			self.gtkwin.fullscreen()
			self.is_fullscreen=True
		
		print "Fullscreen = ",self.is_fullscreen
	
class GpsConnection:
	def __init__(self):
		self.connection=None

	def connect(self):
		print "GPS: connecting to gpsd",
		try:	
			self.connection=gps.gps()
			print ""
			return True
		except:
			print "failed: ",sys.exc_info()[1]
			return False
		
	def update(self):
		if self.connection==None:
			if self.connect()==False:
				return False
		try:
			self.connection.query('admosy')
			return True
		except:
			self.connection=None
			return False
			
	def get_data(self):
		"""return some gps data as dict
		"""
		data={'latitude' : self.connection.fix.latitude,
			'longitude' : self.connection.fix.longitude,
			'utc' : self.connection.utc, 
			'status':self.connection.status, 
			'mode':self.connection.fix.mode, 
			'time':self.connection.fix.time,
			'altitude' : self.connection.fix.altitude,
			'eph' : self.connection.fix.eph,
			'epv' : self.connection.fix.epv,
			'ept' : self.connection.fix.ept,
			'speed' : self.connection.fix.speed,
			'climb' : self.connection.fix.climb,
			'satellites':self.connection.satellites
			}
		return data;

if __name__ == '__main__':
	GGPS()
	try:
		gtk.main()
	except KeyboardInterrupt:
		print "Exit"
